#include <stdio.h>
#include <stdlib.h>

#include "atmroute.h"
#include "atmcell.h"
#include "atmnode.h"
#include "atmnetwork.h"
#include "atmagent.h"

/*
 * This class identifies that "ns node atm-network $route" is actually 
 * creating an object of type ATMNetwork
 */
static class ATMNetworkMatcher : public Matcher {
public:
	ATMNetworkMatcher() : Matcher("atm-network") {}
	TclObject* match(const char* id) {
		return (new ATMNetwork());
	}
} matcher_atm_network;


ATMNetwork::ATMNetwork()
{
	map=new int[64]; //Right now 64 is more than enough. 
			//It will double whenever it gets full
	earliest_start=new double[64];
	for(int i=0;i<64;i++)
	{
		map[i]=-1;
		earliest_start[i]=0.0;
	}
	curr_nodes=0;		
	max_nodes=32;
	nodes=new ATMNode*[max_nodes];
	link_table=new ATMLink**[max_nodes];
	for(int i=0;i<max_nodes;i++)
		link_table[i]=new ATMLink*[max_nodes];
	routelogic=NULL;
}
ATMNetwork::~ATMNetwork()
{
}

/* 
 * Given the name of the object, returns the index of the node in the network
 * The name is obtained from the name() function of the TclObject class.
 */
int ATMNetwork::lookup_id(const char * id)
{
        for(int i=0;i<curr_nodes;i++)
                if(!strcmp(nodes[i]->name(),id))
                        return i;
	return -1;
}

/*
 * This function records that which TCP Node is attached to which ATM Node
 * It uses the TclObject::name() function to identify objects. Returns
 * whether it succeeded or not.
 */
int ATMNetwork::attach(Node * t,ATMNode* a)
{
        int index=lookup_id(a->name());
        if(index<0)
        {
		Tcl & tcl=Tcl::instance();
		tcl.result("attach: Node not found in this atm network.");
                return (TCL_ERROR);

        }
        map[t->addr()]=index;
	a->add_dummy_link(t->addr());
	return (TCL_OK);
}

/*
 * This function is called from the Tcl input and by matching the user command
 * performs that actions associated with it. It returns whether it succeeded or
 * not
 */
int ATMNetwork::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
        if (argc == 3) 
	{
		//  Associate a routing logic with this network

                if (strcmp(argv[1], "route") == 0) {
			routelogic=(ATMRouteLogic*)lookup(argv[2]);
			if(!routelogic)
	                {
        	                tcl.result("ATMNetwork route: Unable to find object");
                	        return (TCL_ERROR);
                 	}
        		return (TCL_OK);
                }
		//Lookup for the node and add it to the network.
		if (strcmp(argv[1], "add-node") == 0) {
			ATMNode* temp=(ATMNode*)lookup(argv[2]);
			if(!temp)
	                {
        	                tcl.result("ATMNetwork add-node: Unable to find object");
                	        return (TCL_ERROR);
                 	}
			add_node(temp);
                        return (TCL_OK);
                }
	}
	//If the ATMNetwork can not handle this command then let its parent
	//handle it.
	return (Node::command(argc,argv));
}

/*
 * This function is called via the command function and adds the node to the
 * network. If the nodes reach their maximum then simply double the array
 */
void ATMNetwork::add_node(ATMNode * node)
{
	if(curr_nodes>=max_nodes)
	{
		max_nodes*=2;
		ATMNode ** t_nodes=new ATMNode*[max_nodes];
		int i=0;
		for(;i<curr_nodes;i++)
		{
			t_nodes[i]=nodes[i];
			t_nodes[i]->init_aal(max_nodes);
		}
		for(;i<max_nodes;i++)
		{
			t_nodes[i]=(0);
			t_nodes[i]->init_aal(max_nodes);
		}
		nodes=t_nodes;
	}
	node->init_aal(max_nodes);	//initialize the aal layer of the node
	node->addr(curr_nodes);		//Set its address to be the current 
					//sequence number
	node->network(this);		//Node records that it belongs to this
					//Network
	nodes[curr_nodes++]=node;	//add the node to the array
	routelogic->curr_nodes(curr_nodes);   //Set route logic number of nodes
	routelogic->Nodes(nodes);
	routelogic->Links(link_table);
}

/*
 * This function handles the events particularly generated for this network.
 * If a packet comes it breaks it up and passes is on to the appropriate atm
 * node. When the packet recovers it forwards it back to the TCP network.
 */
void ATMNetwork::handle(Event* e)
{
        if (route_ == 0) {
                fprintf(stderr, "ns: can't run without routes\n");
                exit(1);
        }
	Scheduler& s = Scheduler::instance();
        Packet* p = (Packet*)e;

	//When a packet comes to the network for the first time, its ldst_ will
	//be pointing to this node. So we break it up into 47 bytes each but
	//transmit at 53 bytes to the atmnode. At the same time we change the 
	//ldst_ to be something other than the current node address. When it
	//arrives again, its ldst_ will be changed and then it is handled just
	//like any other TCP node.

	if(addr_ != p->ldst_) //Recovered Packet
	{
		p->ldst_=addr_;
		ofstream fout("output",ios::app);
		fout<<"\n"<<s.clock()<<"\tRecovering Pkt\t"<<(ATMCell*)p;
			//->seqno_<<"\t"
			//<<p->src_<<" -> "<<p->dst_;
		fout.close();
	}
	else 	//Break a fresh packet from the network
	{
		p->ldst_= addr_+1;
		ATMCell * cell=(ATMCell*)p;
		cell->source_=map[p->lsrc_];	//map gives the node from which
					//the cell should be sent
		cell->dest_=map[(route_[p->dst_])->dst()];	//The outgoing
				//TCP node from map gives the atm destination 
					
                cell->total_cells_=cell->size_/47;
		cell->priority_=0;		//Set cell priority to 0 since
						//UBR cell
		cell->queued_=0;		//Currently Not queued

                if(cell->size_%47 > 0)		//If part of packet left
                        cell->total_cells_ ++;

                //This link is the input link!!!
                ATMLink * link = (ATMLink*)route_[p->lsrc_];

		//Transmission time (based on 53 bytes -> Header attached)
                double txtime = 424. / link->bandwidth();
                ATMCell * newcell;

		//Based on the routing strategy establish routing tables
		//and return any delay associated with establishing a route
		double delay=routelogic->establish_route(cell->source_,
					cell->dest_);

		//This gives the time the packet must wait before it can be
		//handled. This depends upon multiple packet arrivals from 
		//different links.
                double wait=earliest_start[p->lsrc_]-s.clock()+delay;
                if(wait<0)
                        wait=0;
                for (int i=0;i<cell->total_cells_;i++)
                {
                        newcell=new ATMCell;

                        //Allocate new memory for the cell.
                        *newcell=*cell;
                        newcell->atmseq_no_=i+1;
			//Schedule the event at the cell source
                        s.schedule(nodes[cell->source_] , newcell, wait+i*txtime);
                }

		//Reset the value of earliest start
                earliest_start[p->lsrc_]=wait + s.clock() + i*txtime;
		ofstream fout("output",ios::app);
		fout<<"\n"<<s.clock()<<"\tBreaking\t"<<(ATMCell*)p;
			/*addr_<<"\t"
			<<p->seqno_<<"\t"<<p->src_<<" -> "<<p->dst_;*/
		fout.close();
		return;
        }

	//Handle just like a TCP node
        if (p->dst_ & MCASTADDR)
                multicast_forward(p);
        else
                unicast_forward(p);
}

/*
 * Add the link to the list of all links in a network. You can not have multiple
 * links between 2 nodes.
 */

void ATMNetwork::add_link(ATMLink* link, int src,int dst)
{
	link_table[src][dst]=link;
}

int ATMNetwork::CAC(QOS * qos, int src, int dst)
{
	double req_bw= (qos->PCR) * 424.0;
	routelogic->establish_route(src,dst);
	int cur_n=src;
	while(cur_n!=dst)
	{
		ATMLink* lk=(nodes[cur_n])->get_route(dst);
		if(!lk)
			return 0;
		if(lk->available_bw()<req_bw)
			return 0;
		cur_n=(lk->neighbor())->addr();
	}
	
	cur_n=src;
	while(cur_n!=dst)
	{
		ATMLink* lk=(nodes[cur_n])->get_route(dst);
		lk->allocate_bw(req_bw);
		cur_n=(lk->neighbor())->addr();
	}
	return 1;
}

void ATMNetwork::deallocate_resources(QOS * qos, int src, int dst)
{
	double alloc_bw= (qos->PCR) * 424.0;
	int cur_n=src;
	while(cur_n!=dst)
	{
		ATMLink* lk=(nodes[cur_n])->get_route(dst);
		lk->deallocate_bw(alloc_bw);
		cur_n=(lk->neighbor())->addr();
	}
}
