#include <stdio.h>
#include <stdlib.h>

#include "packet.h"
#include "node.h"
#include "link.h"
#include "agent.h"
#include "atmswitch.h"
#include "atmnetwork.h"


ATMSwitch::ATMSwitch()
{
	buffer=new ATMSwitchBuffer;
	link_int("ns_switch", "buffer-size",&buffer_size_, 0);
}

void ATMSwitch::init_aal(int tn)
{
	ATMNode::init_aal(tn);
}

/*
 * Add a link to the switch. Also update the buffer sizes allocated to each
 * port.
 */
void ATMSwitch::add_link(ATMLink* link)
{
	ATMNode::add_link(link);
	set_buffer_limit();

	if(buffer->qsize_)
	{
		delete buffer->qsize_;
		delete buffer->exit_time;
	}
	buffer->qsize_=new int[no_links_ + dummy_count];
	buffer->exit_time=new double[no_links_ + dummy_count];
	for(int i=0;i<no_links_ + dummy_count;i++)
	{
		buffer->qsize_[i]=0;
		buffer->exit_time[i]=0.0;
	}
	
}
/*
 * $switch discard-algo $discard -> sets discard algorithm
 * $switch buffer-size $size -> sets buffer-size
 * $switch loss -> Gives the cells lost
 * $switch passed -> Gives the cells passed
 * $switch qlen $port -> Gives the current queue length of the buffer
 * $switch reset -> set all statistics to 0
 */

int ATMSwitch::command(int argc, const char*const* argv)            
{
	Tcl& tcl = Tcl::instance();
	if (argc==2)
	{
		if (strcmp(argv[1], "loss") == 0) {
			tcl.resultf("%d" , no_of_cells_lost);
			return (TCL_OK);
		}
		else if (strcmp(argv[1], "passed") == 0) {
			tcl.resultf("%d" , no_of_cells_passed);
			return (TCL_OK);
		}
		else if (strcmp(argv[1], "reset") == 0) {
			no_of_cells_lost=no_of_cells_passed=0;
			return (TCL_OK);
		}
	}
	else if(argc==3)
	{
		if (strcmp(argv[1], "qlen") == 0) {
			int port=atoi(argv[2]);
			if(port >= no_links_ + dummy_count)
			{
				tcl.resultf("switch qlen: no such port");
				return (TCL_ERROR);
			}
			tcl.resultf("%d" , buffer->qsize_[port]);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "discard-algo") == 0) {
			discard_=(DiscardAlgorithm*)lookup(argv[2]);
                        if(!discard_)
                        {
                                tcl.result("ATMSwitch : Unable to find object");
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
		else if (strcmp(argv[1], "buffer-size") == 0) {
			buffer_size_=atoi(argv[2]);
                        if(buffer_size_<=0)
                        {
                                tcl.result("ATMSwitch : Illegal Buffer size");
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
	}
	return (NsObject::command(argc, argv));
}

/*
 * Handle the cell
 */
void ATMSwitch::handle(Event* x)
{
        ATMCell* cell=(ATMCell*)x;
        Scheduler& s = Scheduler::instance();

        //if I am the destination and the cell is queued then handover to the 
	//appropriate aal layer else if I am the destination then queue with
	//the appropriate dummy port.
        //If the packet is complete then generate an immediate event in the
        //atm network to return the packet. Also if I am the source the source
        //and the destination then return the packet.
	//If I am not the destination then put in queue. If already queued
	//then put on the next link. Before enqueing check whether it 
	//should be discarded or not.

        if(cell->dest_ == addr_ && cell->queued_)
        {
		int b_port= get_dummy_port(network_->dummy_addr(cell->dst_));
		deque(cell,b_port);
                if(cell->source_ == addr_)
                        s.schedule(network_, x , 0);
                if(aal[cell->source_].Receive_Cell(cell))
                        s.schedule(network_, x , 0);
                else
                        delete cell;    //Free memory since the last cell in
                                //this packet would clear it up.
		return;
        }
        else if(cell->dest_ == addr_ )
	{
		int b_port=get_dummy_port(network_->dummy_addr(cell->dst_));
                if(!(discard_)->check_discard(buffer, cell ,b_port))
		{
			enque(cell,b_port);
			double slot=
	376.0/(((ATMLink*)((network_->route())[cell->dst_]))->bandwidth());
			(buffer->exit_time[b_port])+=slot; 
        		s.schedule(this, cell,
				buffer->exit_time[b_port]-s.clock());
		}
		else
		{
                        no_of_cells_lost++;
                        ofstream fout("output",ios::app);
                        fout<<"\n"<<s.clock() <<"\tDropping Cell\t"<<cell;
			fout<<"\tQ("<<buffer->qsize_[b_port]<<","
				<<buffer->qlimit_<<")";
                        fout.close();
                }
		return;
	}
        else if(!cell->queued_)
        {
                int port=cell->dest_;
                int b_port=port_table[((links_[port])->neighbor())->addr()];
                if(!(discard_)->check_discard(buffer, cell ,b_port))
		{
        		int port=cell->dest_;
        		int b_port=port_table[((links_[port])->neighbor())->addr()];
			enque(cell,b_port);
			double slot=((424.0/(links_[port])->bandwidth()));
			(buffer->exit_time[b_port])+= slot;
        		s.schedule(this, cell,
				buffer->exit_time[b_port]-s.clock());

                }
                else {
                        no_of_cells_lost++;
                        ofstream fout("output",ios::app);
                        fout<<"\n"<<s.clock() <<"\tDropping Cell\t"<<cell;
			fout<<"\tQ("<<buffer->qsize_[b_port]<<","
				<<buffer->qlimit_<<")";
                        fout.close();
                }
        }
        else 
	{
        	ATMLink * link=links_[cell->dest_];
        	int b_port=port_table[(link->neighbor())->addr()];
		deque(cell,b_port);
        	s.schedule(link, cell, 0);
        }
}
/*
 * Insert the cell in the appropriate queue and update the stats
 */
void ATMSwitch::enque(ATMCell* cell,int b_port)
{
        Scheduler& s = Scheduler::instance();
	if(buffer->exit_time[b_port] < s.clock())
		buffer->exit_time[b_port]=s.clock();
        double length=++(buffer->qsize_[b_port]);
        (buffer->total_q_)++;
        no_of_cells_passed++;
        cell->queued_=1;
        ofstream fout("output",ios::app);
        fout<<"\n"<<s.clock()<<"\tEnqueuing Cell\t"<<cell;
	fout<<"\tQ("<<buffer->qsize_[b_port]<<","
	<<buffer->qlimit_<<")";
        fout.close();
}
/*
 * Deque the cell from the appropriate queue and update the stats
 */
void ATMSwitch::deque(ATMCell* cell,int b_port)
{
        Scheduler& s = Scheduler::instance();
        int port=cell->dest_;
        cell->queued_=0;
        (buffer->qsize_[b_port]) --;
        (buffer->total_q_)--;

        ofstream fout("output",ios::app);
        fout<<"\n"<<s.clock()<<"\tDequeuing Cell\t"<<cell;
	fout<<"\tQ("<<buffer->qsize_[b_port]<<","
		<<buffer->qlimit_<<")";
        fout.close();
}
